/*******************************************************************************
Chapter 11

Depth first search with accumulator (findPathDFacc) and hole (findPathBFholes).
Iterative depth first search (findPathDFitrV1, V2 & V3)
Breadth first search with accumulator (findPathBF) and hole (findPathBFholes).
Backward depth first (findPathBackwardDF).
Bidirectional with iterative deepening (findPathBidirectional).

Example files: f11_4.pl, f11_maze.pl
********************************************************************************
********************************************************************************
Depth first search using Accumulator (Figure 11.7)

findPathDFacc(StartNode,  % the starting node
              Path)       % The depth first path
*******************************************************************************/

findPathDFacc(StartNode, Path) :-
    depthFirstAcc(StartNode, [], RP), % Second argument is the path accumulator
    reverse(RP,Path).                 % Put path in the forward order

/*******************************************************************************
depthFirstAcc(Node    % First node in the rest of the path
              Acc     % path accumulator
              Path    % [Node|Acc] = Path
*******************************************************************************/

depthFirstAcc(Node, Acc, [Node|Acc]) :-
    goal(Node).

depthFirstAcc(Node, Acc, Path) :-
    s(Node, SuccNode),
    \+ member(SuccNode, Acc),     % prevent a cycle
    depthFirstAcc(SuccNode, [Node|Acc], Path).

/*******************************************************************************
********************************************************************************
Depth first search using holes
*******************************************************************************/

findPathDFhole(StartNode, Path) :-
    depthFirstHole(StartNode, [], Path).

/*******************************************************************************
depthFirstHole(Node    % First node in the rest of the path
               Acc     % path accumulator
               Path)   % [Node|Acc] = Path
*******************************************************************************/

depthFirstHole(Node, _, [Node]) :- goal(Node).

    % The hole is the argument Path which is filled in as the computation
    % proceeds, thereby maintaining a queue.

depthFirstHole(Node, AccPath, [Node | Path]) :-
    s(Node, SuccNode),
    \+ member(SuccNode, AccPath ),
    depthFirstHole(SuccNode, [Node|AccPath], Path).

/*******************************************************************************
********************************************************************************
Figure 11.8 Depth-limited depth-first search
*******************************************************************************/

maxDepthFirst(Node, [Node], _) :- goal(Node).

maxDepthFirst(Node, [Node|RestOfPath], MaxDepth) :-
    MaxDepth > 1,
    s(Node, SuccNode),
    DepthLeft is MaxDepth - 1,
    maxDepthFirst(SuccNode, RestOfPath, DepthLeft).

/*******************************************************************************
********************************************************************************
Iterative depth first search uses Figure 11.8
Equivalent to breadth first search where, instead of maintaining the paths,
the paths are recomputed.

Exercises: Develop V1, V2, find and discuss problem with V3.
Could ask what problems does each solution present.  What can be done to
fix those problems.  What are the advantages and disadvantages of the fix.
*******************************************************************************/

findPathDFitrV1(StartNode, Path, MaxDepth) :-
    asserta(dfitrcount(0)),
    repeat,
        retract(dfitrcount(N)),
        N1 is N + 1,
% write('Paths of max depth '), write(N1), nl,
        asserta(dfitrcount(N1)),
        (maxDepthFirst(StartNode, Path, N1)
        ; N1 >= MaxDepth,
          ! ,
          retract(dfitrcount(_))).

findPathDFitrV2(StartNode, Path, MaxDepth) :-
    findPathDFcount(StartNode, Path, 0, MaxDepth).

findPathDFcount(StartNode, Path, Depth, MaxDepth) :-
    Depth < MaxDepth,
    NewDepth is Depth + 1,
% write('Paths of max depth '), write(NewDepth), nl,
    ( maxDepthFirst(StartNode, Path, NewDepth)
    ; findPathDFcount(StartNode, Path, NewDepth, MaxDepth)).

/*******************************************************************************
From p269
There is a problem with path, as it keeps reprocessing the tree over and over.
How can we stop it?
Solution is only good if the goal can be reached in a reasonable time.
*******************************************************************************/

findPathDFitrV3(StartNode, Path) :-
    path(StartNode, GoalNode, RP),
    goal(GoalNode),
    reverse(RP,Path).

% path(Node1, Node2, Path)
% Path is an acyclic reverse path between Node1 and Node2.
% Using Fig 11.4 use path(a, Last, Path) to find alternative paths.

path(Node, Node, [Node]).                    % Single node path

path(FirstNode, LastNode, [LastNode|Path]) :-
      % Recursive call extends root path by length one.
    path(FirstNode, OneButLastNode, Path),   % Path up to OneButLastNode

% Through backtracking extend root path all possible ways by one node.

    s(OneButLastNode, LastNode),             % Last step
    \+ member(LastNode, Path).               % No cycles

/*******************************************************************************
********************************************************************************
Breadth first search uses an accumulator.
*******************************************************************************/
% Figure 11.10 edited

findPathBFacc(StartNode, Path) :-
    breadthFirst([[StartNode]], RP),
    reverse(RP,Path).

/*******************
breadthFirst(CandidatePaths,  % All paths, yet to be extended.
             Acc )            % Path accumulator
*******************/

    % Base case, do not care about rest of candidate paths as the first candidate
    % path is a path to the goal.

breadthFirst([[Node | Path] | _], [Node | Path]) :-
    goal(Node).

    % Recursive case.  Remove first path from candidate list and extend
    % Candidate paths with all the extensions of the first path.

breadthFirst([FirstPath | RestPaths], Path) :-
    extend(FirstPath, ExtendedPaths),
    append(RestPaths, ExtendedPaths, NewPaths),   % Expensive operation
    breadthFirst(NewPaths, Path).

% Why is the second arg in findall considered to be a predicate?

extend([Node | Path], ExtendedPaths) :-
    findall([NewNode, Node | Path],
        (s(Node, NewNode), \+member(NewNode, [Node | Path])),
        ExtendedPaths).

/*******************************************************************************
Breadth first search using difference lists to extend the list of paths.
An accumulator is used to extend each path, which is why the paths
need to be reversed on output.
*******************************************************************************/
% Figure 11.11 edited

findPathBFdl(StartNode, Path) :-
    breadthFirstDL([[StartNode] | Hole] - Hole, RP),
    reverse(RP,Path).

/*******************
breadthFirstDL(CandidatePaths,  % All paths, yet to be extended.
               Acc )            % Path accumulator
*******************/

    % Base case, do not care about the hole value as the first candidate path
    % is a path to the goal.

breadthFirstDL([[Node | Path] | _] - _, [Node | Path]) :-
    goal(Node).

    % Recursive case.  Remove first path from candidate list and extend
    % Candidate paths with all the extensions of the first path.

breadthFirstDL([FirstPath | RestPaths] - Hole, Path) :-
%nl, write(RestPaths),
    extend(FirstPath, ExtendedPaths),
    append(ExtendedPaths, NewHole, Hole),   % Add new path at end

%nl, write(FirstPath), write(' :: '), write(ExtendedPaths),
%nl, write(RestPaths) , write(' :: '), write(NewHole),
%nl, write(Hole), nl, nl,

    RestPaths \== NewHole,     % Cannot be identical structures
    breadthFirstDL(RestPaths - NewHole, Path).

/*******************************************************************************
Note: There is no magic about the '-', change to '+' and it still works.
I.e. use any infix binary operator.
*******************************************************************************/

/*******************************************************************************
********************************************************************************
Backward depth first search
*******************************************************************************/

findPathBackwardDF(StartNode, Path) :-
    goal(Goal),
    backward(StartNode,[Goal], Path).

/*******************
backward(StartNode,   % Initial state
         Acc,         % Path accumulator
         Path )       % Reverse path
*******************/

    % Base case when first node in the accumulator is the start node,
    % then the the accumulator is the reverse path.

backward(StartNode, [StartNode | Path], [StartNode | Path]).

    % Recursive case, find a predecessor to the node at the end of the
    % path in the accumulator, extend the accumulator, and repeat with
    % the new accumulator.

backward(StartNode, [SuccNode | LastOfPath], Path) :-
    s(PrevNode, SuccNode),
    backward(StartNode, [PrevNode, SuccNode | LastOfPath], Path).

/*******************************************************************************
********************************************************************************
Bidirectional search with iterative deeping
Goes forever after last path has been found.

Program text (edited) from Exercise 11.12 in Bratko
Example: findPathBidirectional(a, Path, MeetingNode)
*******************************************************************************/

findPathBidirectional(StartNode, Path, MeetingNode) :-

    similar_length(ToPath, FromPath),    % Generate two similar length lists.
                                         % abs(#ToPath - #FromPath) <= 1
             % Effect is to have iterative deepening on the forward and
             % backward depth first searchs.

    goal(Goal),                          % Need a goal for backward search

    backwardPath(Goal, RFP, MeetingNode),  % RFP = path(Goal .. MeetingNode)
                                           
    forwardPath(StartNode, ToPath, MeetingNode),  % ToPath = path(StartNode
                                                  %              .. MeetingNode)
    reverse(RFP, [_ | FromPath]),   % RFP = Goal .. MeetingNode
                                    % reverse(RFP) = MeetingNode .. Goal
                                    % MeetingGoal is already in ToPath
                                    % Do not include twice.
                                    % FromPath = tail(reverse(RFP))
 
    append(ToPath, FromPath, Path). % Path = ToPath || FromPath

% Forward path uses holes, and assumes that there are no cycles, as in the
% example tree.

forwardPath(Node, [Node], Node).

forwardPath(First, [First | Rest], Last) :-
    s(First, Second),
    forwardPath(Second, Rest, Last).

% Backward path uses holes, and assumes that there are no cycles, as in the
% example tree.

backwardPath(Node, [Node], Node).

backwardPath(First, [First | Rest], Last) :-
    s(Second, First),
    backwardPath(Second, Rest, Last).

% Generate List1 and List2 such that
% #List1 = #List2 or #List1 + 1 = #List2

similar_length(List1, List2) :-
    equal_length(List2, List),           % Generate equal length lists with
                                         % different sets of variables
    (List1 = List ; List1 = [_ | List]). % List1 is same length as List2
                                         % or one longer.

equal_length([],[]).

equal_length([_|RestList1], [_|RestList2]) :-
    equal_length(RestList1, RestList2).
